/*
** $Header: /u/spert/src/fltvec/nv/RCS/vecred.c,v 1.10 1996/06/10 20:28:30 davidj Exp $
**
** Vector reduction operations - the quick and easy way
*/

#include <fltv-config.h>
#include <assert.h>
#include <stddef.h>
/* Note - using Sun SC3.0 acc, files including <assert.h> must also include
   <stdio.h> */
#include <stdio.h>
#include <math.h>
#include <fltv-defs.h>
#include <fltv-scansrc.h>

/*****************************************************************
** Find the maximum and minimum of a vector
*****************************************************************/

#ifdef DECL_nv_maxmin_vf_ff
void nv_maxmin_vf_ff(size_t n, const float *vec, float *maxp, float *minp);
#endif

#ifdef DEFN_nv_maxmin_vf_ff
void
nv_maxmin_vf_ff(size_t n, const float *vec, float *maxp, float *minp)
{
    if (n>0)
    {
        float max = vec[0];
        float min = vec[0];
        size_t i;
        for (i=1; i<n; i++)
        {
            const float elem = vec[i];
            if (elem > max)
                max = elem;
            else if (elem < min)
                min = elem;
        }

        *maxp = max;
        *minp = min;
    }
}
#endif /* #ifdef DEFN_nv_maxmin_vf_ff */

/****************************************************************/
/* Find the index of them max/min/absmax/absmin of a vector */
/****************************************************************/

#ifdef DECL_nv_imax_vf_u
size_t nv_imax_vf_u(size_t n, const float* v);
#endif
#ifdef DECL_nv_imin_vf_u
size_t nv_imin_vf_u(size_t n, const float* v);
#endif
#ifdef DECL_nv_iabsmax_vf_u
size_t nv_iabsmax_vf_u(size_t n, const float* v);
#endif
#ifdef DECL_nv_iabsmin_vf_u
size_t nv_iabsmin_vf_u(size_t n, const float* v);
#endif

#ifdef DEFN_nv_imax_vf_u
#define NV_MAXMINOP_VF_U nv_imax_vf_u
#define CMP(a,b) (a>b)
#define ABS(a) (a)
#endif

#ifdef DEFN_nv_imin_vf_u
#define NV_MAXMINOP_VF_U nv_imin_vf_u
#define CMP(a,b) (a<b)
#define ABS(a) (a)
#endif

#ifdef DEFN_nv_iabsmax_vf_u
#define NV_MAXMINOP_VF_U nv_iabsmax_vf_u
#define CMP(a,b) (a>b)
#define ABS(a) fabs(a)
#endif

#ifdef DEFN_nv_iabsmin_vf_u
#define NV_MAXMINOP_VF_U nv_iabsmin_vf_u
#define CMP(a,b) (a<b)
#define ABS(a) fabs(a)
#endif

#if defined(DEFN_nv_imax_vf_u) || defined(DEFN_nv_iabsmax_vf_u) \
    || defined(DEFN_nv_imin_vf_u) || defined(DEFN_nv_iabsmin_vf_u)

size_t
NV_MAXMINOP_VF_U(size_t n, const float* v)
{
    float best_val = ABS(v[0]);
    float index = 0;
    int nan = isnan_f_b(best_val); /* 1 if we have seen a nan */
    size_t i;

    assert(n>0);
    for (i=1; i<n; i++)
    {
        const float val = ABS(v[i]);
	nan |= isnan_f_b(val);
        if (CMP(val, best_val))
        {
            best_val = val;
            index = i;
        }
    }
    if (nan)
	return(FLTV_SIZET_BAD);
    else
	return index;
}
#endif /* #if defined(DEFN_imax_vf_u) || .... */

/****************************************************************
** Sum of all elements of a vector
****************************************************************/

#ifdef DECL_nv_sum_vf_f
float nv_sum_vf_f(const size_t len, const float* in);
#endif

#ifdef DEFN_nv_sum_vf_f
float
nv_sum_vf_f(const size_t len, const float* in)
{
    float sum = 0.0f;
    size_t i;

    for (i=0; i<len; i++)
	sum += in[i];

    return(sum);
}
#endif

/****************************************************************/
/* Dot product */
/****************************************************************/

#ifdef DECL_nv_mulsum_vfvf_f
float nv_mulsum_vfvf_f(const size_t n, const float* avec, const float* bvec);
#endif
#ifdef DECL_nv_mulsum_vdvd_d
double nv_mulsum_vdvd_d(const size_t n, const double* avec,
			const double* bvec);
#endif


#if defined(DEFN_nv_mulsum_vfvf_f) || defined(DEFN_nv_mulsum_vdvd_d)
#ifdef DEFN_nv_mulsum_vfvf_f
#define TYPE float
#define NAME nv_mulsum_vfvf_f
#endif
#ifdef DEFN_nv_mulsum_vdvd_d
#define TYPE double
#define NAME nv_mulsum_vdvd_d
#endif

TYPE
NAME(const size_t n, const TYPE* avec, const TYPE* bvec)
{
    register TYPE sum = 0.0f;
    size_t i;

    for (i=0; i<n; i++)
        sum += avec[i] * bvec[i];

    return sum;
}

#endif /* defined(DEFN_nv_mulsum_vfvf_f) || defined(DEFN_nv_mulsum_vdvd_d) */
