/*
   $Header: /u/spert/src/fltvec/testsuite/RCS/camfile_test.c,v 1.2 1996/04/04 01:42:57 davidj Exp $

   Test element wise vector routines of the form op_vfvf_vf.
*/

#include <stddef.h>
#include <stdlib.h>

#include "fltvec_test.h"



void
test_pre8(void)
{
    enum {LEN = 8 };
    fltv_uint8_t testdata[LEN] = { 0, 1, 2, 0x34, 0x7f, 0x80, 0x67, 0xff };
    fltv_int8_t* in;
    float* temp;
    fltv_int8_t* res;
    int i;

    rtst_start("pre8");

    in = rtst_padvec_clone_vi8(LEN, (rtst_int8*) testdata);
    temp = rtst_padvec_new_vf(LEN);
    res = rtst_padvec_new_vi8(LEN);

    pre8toh_vi8_vf(LEN, in, temp);
    htopre8_vf_vi8(LEN, temp, (fltv_int8_t*) res);
    rtst_checkeq_vi8vi8(LEN, testdata, res);
    for (i=0; i<LEN; i++)
    {
	rtst_assert(pre8toh_i8_f(testdata[i])==temp[i]);
	rtst_assert(htopre8_f_i8(temp[i])==res[i]);
    }
    rtst_padvec_del_vi8(in);
    rtst_padvec_del_vf(temp);
    rtst_padvec_del_vi8(res);
    rtst_passed();
}

void
test_lna8(void)
{
    enum {LEN = 8 };
    fltv_uint8_t testdata[LEN] = { 0, 1, 2, 0x34, 0x7f, 0x80, 0x67, 0xff };
    fltv_int8_t* in;
    float* temp;
    fltv_int8_t* res;
    int i;

    rtst_start("lna8");

    in = rtst_padvec_clone_vi8(LEN, (rtst_int8*) testdata);
    temp = rtst_padvec_new_vf(LEN);
    res = rtst_padvec_new_vi8(LEN);

    lna8toh_vi8_vf(LEN, in, temp);
    htolna8_vf_vi8(LEN, temp, res);
    rtst_checkeq_vi8vi8(LEN, testdata, res);
    for (i=0; i<LEN; i++)
    {
	rtst_assert(lna8toh_i8_f(testdata[i])==temp[i]);
	rtst_assert(htolna8_f_i8(temp[i])==res[i]);
    }

    rtst_padvec_del_vi8(in);
    rtst_padvec_del_vf(temp);
    rtst_padvec_del_vi8(res);
    rtst_passed();
}


int
main(int argc, char *argv[])
{
    rtst_args(argc, argv);

    test_lna8();
    test_pre8();
    rtst_exit();

    return 0;
}
