/*
** $Header: /u/spert/src/fltvec/testsuite/RCS/copy_sm_test.c,v 1.3 1996/03/22 03:57:39 davidj Exp $
**
** Test strided matrix copy routines
*/

#include <stddef.h>
#include <stdlib.h>

#include "fltvec_test.h"

typedef void (*rtn_t)(size_t rows, size_t cols, size_t stride,
			      const float* from, float* to);

/* Performs one test given input values and expected output - mf to smf. */

void
one_test_mf_smf(rtn_t rtn, size_t rows, size_t cols, size_t stride,
		const float* from, const float* to)
{
    size_t i;

    /* Create padded copies of inputs. */
    const size_t from_size = rows * cols;
    const size_t to_size = rows * stride;
    float* c_from = rtst_padvec_clone_vf(from_size, from);
    float* c_to = rtst_padvec_new_vf(to_size);
    for (i=0; i<to_size; i++)
	c_to[i] = 0.0f;

    /* Run routine. */
    rtn(rows, cols, stride, c_from, c_to);

    /* Check result. */
    rtst_checkeq_vfvf(to_size, c_to, to);

    /* Check input not corrupted. */
    rtst_checkeq_vfvf(from_size, c_from, from);

    /* Free and check padding. */
    rtst_padvec_del_vf(c_from);
    rtst_padvec_del_vf(c_to);
}


/* Performs one test given input values and expected output - smf to mf. */

void
one_test_smf_mf(rtn_t rtn, size_t rows, size_t cols, size_t stride,
		const float* from, const float* to)
{
    size_t i;

    /* Create padded copies of inputs. */
    const size_t from_size = rows * stride;
    const size_t to_size = rows * cols;
    float* c_from = rtst_padvec_clone_vf(from_size, from);
    float* c_to = rtst_padvec_new_vf(to_size);
    for (i=0; i<to_size; i++)
	c_to[i] = 0.0f;

    /* Run routine. */
    rtn(rows, cols, stride, c_from, c_to);

    /* Check result. */
    rtst_checkeq_vfvf(to_size, c_to, to);

    /* Check input not corrupted. */
    rtst_checkeq_vfvf(from_size, c_from, from);

    /* Free and check padding. */
    rtst_padvec_del_vf(c_from);
    rtst_padvec_del_vf(c_to);
}


void
test_mf_smf_1_1s1(rtn_t rtn)
{
    static const float from[] = { 42.0f };
    static const float to[] = { 42.0f };

    one_test_mf_smf(rtn, 1, 1, 1, from, to);
}

void
test_mf_smf_1_1s3(rtn_t rtn)
{
    static const float from[] = { 42.0f };
    static const float to[] = { 42.0f, 0.0f, 0.0f };

    one_test_mf_smf(rtn, 1, 1, 3, from, to);
}

void
test_mf_smf_2_3s5(rtn_t rtn)
{
    static const float from[] =
    {
	1.0f, 2.0f, 3.0f,
	4.0f, 5.0f, 6.0f
    };
    static const float to[] =
    {
	1.0f, 2.0f, 3.0f, 0.0f, 0.0f,
	4.0f, 5.0f, 6.0f, 0.0f, 0.0f
    };

    one_test_mf_smf(rtn, 2, 3, 5, from, to);
}

void
run_mf_smf_directed_tests(rtn_t rtn)
{
    rtst_log("starting directed tests.\n");
    test_mf_smf_1_1s1(rtn);
    test_mf_smf_1_1s3(rtn);
    test_mf_smf_2_3s5(rtn);
    rtst_log("finished directed tests.\n");
}

void
test_smf_mf_1_1s1(rtn_t rtn)
{
    static const float from[] = { 42.0f };
    static const float to[] = { 42.0f };

    one_test_smf_mf(rtn, 1, 1, 1, from, to);
}

void
test_smf_mf_1s3_1(rtn_t rtn)
{
    static const float from[] = { 42.0f, 0.0f, 0.0f };
    static const float to[] = { 42.0f };

    one_test_smf_mf(rtn, 1, 1, 3, from, to);
}

void
test_mf_smf_3s5_2(rtn_t rtn)
{
    static const float from[] =
    {
	1.0f, 2.0f, 3.0f, 0.0f, 0.0f,
	4.0f, 5.0f, 6.0f, 0.0f, 0.0f
    };
    static const float to[] =
    {
	1.0f, 2.0f, 3.0f,
	4.0f, 5.0f, 6.0f
    };

    one_test_smf_mf(rtn, 2, 3, 5, from, to);
}

void
run_smf_mf_directed_tests(rtn_t rtn)
{
    rtst_log("starting directed tests.\n");
    test_smf_mf_1_1s1(rtn);
    test_smf_mf_1s3_1(rtn);
    test_mf_smf_3s5_2(rtn);
    rtst_log("finished directed tests.\n");
}

int
main(int argc, char *argv[])
{
    rtst_args(argc, argv);

    rtst_start("copy_smf_mf");
    rtst_log("Checking nv.\n");
    run_mf_smf_directed_tests(nv_copy_mf_smf);
#ifdef CHECK_WRP
    rtst_log("Checking wrp.\n");
    run_mf_smf_directed_tests(wrp_copy_mf_smf);
#endif    
    rtst_passed();
    rtst_start("copy_mf_smf");
    rtst_log("Checking nv.\n");
    run_smf_mf_directed_tests(nv_copy_smf_mf);
#ifdef CHECK_WRP
    rtst_log("Checking wrp.\n");
    run_smf_mf_directed_tests(wrp_copy_smf_mf);
#endif
    rtst_passed();
    
    rtst_exit();

    return 0;
}
