/*
   $Header: /u/spert/src/fltvec/testsuite/RCS/sumcol_mf_vf_test.c,v 1.2 1996/04/10 22:48:27 davidj Exp $

   Test sumcol_mf_vf routine.
*/

#include <stddef.h>
#include <stdlib.h>

#include "fltvec_test.h"

typedef void (*rtn_t)(size_t rows, size_t cols, const float* mat, float* vec);

/* Performs one test given routine, input values and expected output. */

void
one_test(rtn_t rtn, size_t rows, size_t cols,
         const float* mat, const float* resvec)
{
    /* Create padded copies of inputs. */
    const size_t mat_len = rows * cols;
    float* c_mat = rtst_padvec_clone_vf(mat_len, mat);
    float* c_resvec = rtst_padvec_new_vf(cols);

    /* Run routine. */
    rtn(rows, cols, c_mat, c_resvec);

    /* Check result. */
    rtst_checkeq_vfvf(cols, c_resvec, resvec);

    /* Check input not corrupted. */
    rtst_checkeq_vfvf(mat_len, c_mat, mat);

    /* Free and check padding. */
    rtst_padvec_del_vf(c_mat);
    rtst_padvec_del_vf(c_resvec);
}

void
test_1x1(rtn_t rtn)
{
    static const float m[] = { 12.0f };
    static const float r[] = { 12.0f };

    one_test(rtn, 1, 1, m, r);
}

void
test_3x2(rtn_t rtn)
{
    static const float m[] = {
        1.0f,  3.0f,
        -2.0f,  9.0f,
        3.0f, -3.0f
    };
    static const float r[] = { 2.0f, 9.0f };

    one_test(rtn, 3, 2, m, r);
}

void
run_directed_tests(rtn_t rtn)
{
    rtst_log("starting directed tests.\n");
    test_1x1(rtn);
    test_3x2(rtn);
    rtst_log("finished directed tests.\n");
}

void
run_random_tests(rtn_t rtn)
{
    const int numtests = rtst_numtests;
    const size_t maxlen = rtst_sizetests;
    float *mat = malloc_vf(maxlen*maxlen);
    float *outvec = malloc_vf(maxlen);
    int i;

    rtst_log("starting random tests.\n");
    for (i=0; i<numtests; i++)
    {
        size_t rows = (size_t) rtst_urand_i32i32_i32(1, maxlen);
        size_t cols = (size_t) rtst_urand_i32i32_i32(1, maxlen);

        rtst_urandi_i32i32_vf(rows*cols, -7, +7, mat);
        nv_sumcol_mf_vf(rows, cols, mat, outvec);

        rtst_log("rows=%d cols=%d\n", rows, cols);

        one_test(rtn, rows, cols, mat, outvec);
    }

    free_vf(mat);
    free_vf(outvec);
    rtst_log("finished random tests.\n");
}

int
main(int argc, char *argv[])
{
    rtst_args(argc, argv);

    rtst_start("nv_sumcol_mf_vf");
    run_directed_tests(&nv_sumcol_mf_vf);
    rtst_passed();

    rtst_start("php_sumcol_mf_vf");
    run_directed_tests(php_sumcol_mf_vf);
    run_random_tests(php_sumcol_mf_vf);
    rtst_passed();
    
    rtst_exit();

    return 0;
}
