/*
** $Header: /u/spert/src/fltvec/tor/RCS/tor_conv_gf_gx32s.S,v 1.4 1995/07/05 22:36:29 davidj Exp $
*/

#include "tor/tordef.h"
#include "m_regdef.h"

	.globl tor_conv_gf_gx32s
#ifdef DECL_tor_conv_gf_gx32s 
void tor_conv_gf_gx32s(void); 
#endif

#define vA va0			/* Input A				*/
#define zExp t3			/* Exponent of Z			*/
#define vZ vv2			/* Output Z				*/
#define saturationFlags v0	/* Flags indicating whether each vector	*/
				/*     result element saturated		*/

#define vZExp vt1		/* Exponent of vZ, possibly adjusted	*/
				/*     for subnormal vA			*/
#define zExp2 t4		/* Exponent of Z			*/
#define vRoundingNeeded vtt	/* Flag for conversion is a shift down	*/
				/*     requires rounding		*/
#define vAbsA vt0		/* vA shifted up one bit to remove sign	*/
#define vA2 vt2			/* vA with leading significand bit made	*/
				/*     explicit by overwriting exponent	*/
#define vAbsZ2 vt3		/* Magnitude of vZ if rounding required	*/
#define vShiftAmount vt4	/* Amount to shift down in conversion	*/
#define vShiftAmount2 vt5	/* vShiftAmount minus 2, saturated	*/
				/*     above to 31			*/
#define vAbsZ vt2		/* Magnitude of vZ			*/
#define vANegative vtt		/* Flag for vA negative			*/
#define vNegAbsZ vt3		/* Negative of vAbsZ			*/
#define vShiftedA va2		/* vAbsA shifted for count of leading	*/
				/*     zeros				*/
#define vCountLeadingZerosA vt1	/* Number of leading zeros of vAbsA	*/
#define vCount16 vt2		/* vCountLeadingZerosA & 0x10		*/
#define vCount8 vt3		/* vCountLeadingZerosA & 0x08		*/
#define vLowIndex vt4		/* Index into low zeros-count table	*/
#define vHighIndex vt5		/* Index into high zeros-count table	*/
#define vAExpOffset vt1		/* Normalization adjustment of exponent	*/
				/*     of vA				*/
#define vAExp vt1		/* Exponent of vA			*/
#define vAIsNaNOrInfinity vt2	/* Flag for vA is NaN or infinity	*/
#define vAIsNegInfinity vt3	/* Flag for vA is negative infinity	*/

#define temp0 v1
#define temp1 t5
#define temp2 t6
#define config0 t7
#define config1 t8
#define config2 t9

	.ent tor_conv_gf_gx32s

/*
-------------------------------------------------------------------------------
Routine:  tor_conv_gf_gx32s

Input:
	vlr
	va0	vA
	t3	zExp

Output:
	vv2	vZ
	v0	saturationFlags

Not preserved:
	vtt
	vv3
	va0--va3
	vt0--vt5
	v0--v1
	t4--t9

Function:

The single-precision floating-point vector vA is converted to a 32-bit fixed-
point vector and the result returned in vZ.  The exponent of the vZ vector is
given by zExp.  The result is rounded to nearest/even and saturated as
necessary.  The vlr register controls the elements of vZ actually updated.  The floating-point format conforms to the IEEE Standard for Binary Floating-Point
Arithmetic, and the conversion is performed in a manner consistent with that
standard.  The saturationFlags result indicates for each element in vZ whether
that result saturated.

Exceptions for any of the element-wise conversions are signalled by raising
element-wise flags in the global variable fltflags_invalid32.  A flag in
fltflags_invalid32 is raised if the corresponding element in vA is a NaN.  The
result of attempting to convert a NaN is the maximum positive value 0x7FFFFFFF;
the corresponding flag in saturationFlags is _not_ set.

-------------------------------------------------------------------------------
*/

tor_conv_gf_gx32s:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	va0	vA
|	vt1	vZExp
|	t3	zExp
|	t4	zExp2
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Initially copy zExp to zExp2 and vZExp.  vZExp may later be modified, if vA is
subnormal and if that needs to be taken into account in determining vZ.  In
that case, zExp2 will be set to zero to show that subnormals have already been
dealt with.  (If zExp starts off as zero, subnormals do not need to be taken
into account in determining vZ.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	mov.vs vZExp,zExp
	move zExp2,zExp

repeatAfterNormalization:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	va0	vA
|	vt1	vZExp
|	vt2	vShiftAmount
|	t3	zExp
|	t4	zExp2
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Extract the exponent of vA, adjust it, and subtract from vZExp to form
vShiftAmount, which is the amount to shift down the significand of vA to form
the final result.  (If vShiftAmount is positive, the result must saturate.)
    Zero, subnormal, infinity, and NaN inputs are (mis)treated as though they
were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config0,((1*FX_SHLAM)|FX_AU|FX_BU\
	            |FX_SEPSHAM|(24*FX_SHRAM)|FX_NOROUND|FX_CLIP_PASS)
	li temp0,(0x7F<<24)
	fxsub.vs vShiftAmount,vA,temp0,config0
	ssub.vv vShiftAmount,vZExp,vShiftAmount

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vRoundingNeeded
|	va0	vA
|	vt0	vAbsA
|	vt2	vA2,vAbsZ
|	vt3	vAbsZ2
|	vt4	vShiftAmount
|	vt5	vShiftAmount2
|	t3	zExp
|	t4	zExp2
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Shift up vA by one bit to eliminate its sign.  Set vA2 to be vA with the
leading significant bit made explicit by overwriting the bottom bit of the
exponent with a 1.
    Zero, subnormal, infinity, and NaN inputs continue to be (mis)treated as
though they were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	addu.vv vAbsA,vA,vA
	li temp1,(1<<23)
	or.vs vA2,vA,temp1

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set up a "shift up by 8, shift down variable, round-to-odd" configuration, used
multiple times.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config1,((8*FX_SHLAM)|FX_AU|FX_SHRL|FX_SHRV|FX_RODD|FX_CLIP_PASS)

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vShiftAmount is 8 or more, set vAbsZ2 to be the magnitude of the final
result vZ.  This is accomplished by first subtracting 2 from vShiftAmount and
then saturating to a maximum of 31 to form vShiftAmount2.  vA2 is then shifted
up by 8 bits to extract the significand, then down by vShiftAmount2, and then
shifted down again by the remaining 2 bits and rounded to nearest/even.  If
vShiftAmount is less than 8, vAbsZ2 is trash.  (Or may as well be.)
    (The calculation of vShiftAmount2 below appears differently from that 
stated above.  In effect, 32 is subtracted from vShiftAmount2 before it is used
as a shift amount.  Since only the bottom 5 bits are used as a shift amount,
this has no effect on the operation.  The subtraction of 32 actually occurs
along with the subtraction of 2, so 34 is subtracted altogether.  The amount to
saturate to is then -1 instead of 31.  Subtracting 32 makes the saturation
cheaper.)
    (temp0 is live beyond this set of statements.)
    Zero, subnormal, infinity, and NaN inputs continue to be (mis)treated as
though they were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp2,34
	ssub.vs vShiftAmount2,vShiftAmount,temp2
	li temp0,-1
	cmvgez.vs vShiftAmount2,vShiftAmount2,temp0
	fxadd.vv vAbsZ2,vA2,vShiftAmount2,config1
	li config2,((2*FX_SHLAM)|FX_RNEARESTEVEN|FX_CLIP_PASS)
	fxadd.vv vAbsZ2,vzero,vAbsZ2,config2

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vShiftAmount is in the range 0 to 7 inclusive, set vAbsZ to be the magnitude
of the final result vZ.  This only involves shifting up vA2 to extract the
significand and then shifting down by vShiftAmount.  No rounding is needed.  If
vShiftAmount is outside the range 0 to 7, vAbsZ is trash.  (Or may as well be.)
    Zero, subnormal, infinity, and NaN inputs continue to be (mis)treated as
though they were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	fxadd.vv vAbsZ,vA2,vShiftAmount,config1

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set vAbsZ to the magnitude of the final result.  If vAbsA is zero, vAbsZ is set
to zero.  Else if vShiftAmount is negative, vAbsZ is saturated to 0xFFFFFFFF.
Else if vShiftAmount is 8 or more, vAbsZ is set to vAbsZ2.  Else if
vShiftAmount is in the range 0 to 7 inclusive, vAbsZ is unchanged from above.
    Subnormal, infinity, and NaN inputs continue to be (mis)treated as though
they were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp1,7
	slt.sv vRoundingNeeded,temp1,vShiftAmount
	cmvnez.vv vAbsZ,vRoundingNeeded,vAbsZ2
	cmvltz.vs vAbsZ,vShiftAmount,temp0
	cmveqz.vv vAbsZ,vAbsA,vzero

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vANegative
|	vv0	vZ
|	va0	vA
|	vt0	vAbsA
|	vt2	vAbsZ
|	vt3	vNegAbsZ
|	v0	saturationFlags
|	t3	zExp
|	t4	zExp2
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Form vZ from vAbsZ, conditionally negating according to whether vA is negative,
and saturating to 32 bits.  Record any saturations in saturationFlags.  (If vA
is negative, vZ will saturate exactly when vAbsZ minus 1 saturates.)
    Subnormal, infinity, and NaN inputs continue to be (mis)treated as though
they were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	sltz.vv vANegative,vA
	ctvu zero,vsat
	li config0,(FX_AU|FX_BU|FX_CLIP32)
	fxsub.vv vZ,vAbsZ,vANegative,config0
	cfvu saturationFlags,vsat
	fxsub.vv vNegAbsZ,vzero,vAbsZ,config0
	cmvltz.vv vZ,vA,vNegAbsZ

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vAbsA is nonzero and the exponent of vAbsA is less than 0x80, map vAbsA to
-vAbsA (the negative of vAbsA).  This has the effect in those cases of mapping
the exponent of vAbsA either to (0x100 - exponent) or to (0xFF - exponent).
The mapping cannot overflow; the result of the mapping must be an exponent of
no more than 0xFF.  If the input vA was a subnormal (nonzero with exponent 0),
or if it was an infinity or NaN (exponent 0xFF), the resulting vAbsA must have
an unsigned integer value greater than 0xFEFFFFFF.  If that is the case, jump
to a special case handler.  (If vAbsA is zero, -vAbsA is also zero, so the jump
is not taken.  0xFEFF0000 is used instead of 0xFEFFFFFF because it is cheaper
to load into a register.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config1,(FX_CLIP_PASS|FX_LTZ)
	fxsub.vv vAbsA,vzero,vAbsA,config1
	ctvu zero,vcond
	li temp2,0xFEFF0000
	fltu.sv temp2,vAbsA
	cfvu temp2,vcond
	bnez temp2,specialCases

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Return.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	j ra

specialCases:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vv0	vZ
|	va0	vA
|	vt0	vAbsA
|	v0	saturationFlags
|	t3	zExp
|	t4	zExp2
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Shift up vA by one bit to eliminate its sign.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	addu.vv vAbsA,vA,vA

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Determine whether subnormal inputs need attention.  If the code to handle
subnormals has already been executed once, zExp2 is zero; do not execute it
again.  Alternatively, if the fixed-point result exponent is outside the range
-0x95 to -0x5E inclusive, conversion of a subnormal floating-point input will
either saturate or round to zero.  In that case, subnormal inputs can be
treated as though they were normal without danger, so skip the code to handle
subnormals.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	slt temp0,zExp2,-0x5D
	beqz temp0,subnormalsIrrelevant
	slt temp1,zExp2,-0x95
	bnez temp1,subnormalsIrrelevant

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	va0	vA
|	va2	vShiftedA
|	vt0	vAbsA
|	vt1	vCountLeadingZerosA
|	vt2	vCount16
|	vt3	vCount8
|	vt4	vLowIndex
|	vt5	vHighIndex
|	t3	zExp
|	t4	zExp2
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set vCountLeadingZerosA to the number of leading zeros of vAbsA.
    (NOTE:  This block of code could all be replaced by a single count-leading-
zeros instruction.  See countlz_gi32_gi32.S for more refined commentary.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,(1<<16)
	sltu.vs vCount16,vAbsA,temp0
	li temp1,4
	sllv.vs vCount16,vCount16,temp1
	sllv.vv vShiftedA,vAbsA,vCount16
	li temp2,(1<<24)
	sltu.vs vCount8,vShiftedA,temp2
	li temp0,3
	sllv.vs vCount8,vCount8,temp0
	sllv.vv vShiftedA,vShiftedA,vCount8
	addu.vv vCount8,vCount16,vCount8
	li config0,(FX_SHLZA|FX_A_AND_B|(23*FX_SHLAM)|FX_NOROUND|FX_CLIP_PASS)
	li config1,(FX_SHLV|(28*FX_SHLAM)|FX_NOROUND|FX_CLIP_PASS|FX_GEZ)
	li temp1,(7<<25)
	fxadd.vs vLowIndex,vShiftedA,temp1,config0
	li temp2,0x76554444
	fxadd.sv vCountLeadingZerosA,temp2,vLowIndex,config1
	li temp0,(7<<28)
	addu config2,config0,(3*FX_SHLAM)
	fxadd.vs vHighIndex,vShiftedA,temp0,config2
	li temp1,0xF3221111
	fxadd.sv vCountLeadingZerosA,temp1,vHighIndex,config1
	cmvltz.vv vCountLeadingZerosA,vShiftedA,vzero
	addu.vv vCountLeadingZerosA,vCount8,vCountLeadingZerosA

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	va0	vA
|	vt0	vAbsA
|	vt1	vCountLeadingZerosA,vAExpOffset,vZExp
|	t3	zExp
|	t4	zExp2
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vAbsA is subnormal, normalize it by shifting up.  Set vAExpOffset to the
number of bits shifted (zero if normal, infinity, or NaN).
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp2,7
	subu.vs vAExpOffset,vCountLeadingZerosA,temp2
	cmvltz.vv vAExpOffset,vAExpOffset,vzero
	sllv.vv vAbsA,vAbsA,vAExpOffset

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Form a normalized vA by combining the normalized vAbsA with the sign from the
old vA.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,1
	srlv.vs vAbsA,vAbsA,temp0
	li temp1,0x80000000
	and.vs vA,vA,temp1
	or.vv vA,vA,vAbsA

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Combine vAExpOffset with zExp2 to give vZExp, and set zExp2 to zero to signal
that subnormals have already been normalized.  Start over.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	addu.vs vZExp,vAExpOffset,zExp2
	li zExp2,0
	b repeatAfterNormalization

subnormalsIrrelevant:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vv0	vZ
|	va0	vA
|	vt0	vAbsA
|	vt1	vAExp
|	vt2	vAIsNaNOrInfinity
|	vt3	vAIsNegInfinity
|	v0	saturationFlags
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Extract exponent from vAbsA.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,24
	srlv.vs vAExp,vAbsA,temp0

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set vAIsNaNOrInfinity nonzero exactly if vA is a NaN or infinity.  Set
vAIsNegInfinity nonzero exactly if vA is negative infinity.  If vA is a NaN or
is positive infinity, set vZ to 0x7FFFFFFF; else if vA is negative infinity,
set vZ to -0x80000000.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,0xFF
	seq.vs vAIsNaNOrInfinity,vAExp,temp0
	li temp1,0x80000000
	subu temp2,temp1,1
	cmvnez.vs vZ,vAIsNaNOrInfinity,temp2
	li temp0,0xFF800000
	seq.vs vAIsNegInfinity,vA,temp0
	cmvnez.vs vZ,vAIsNegInfinity,temp1

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vA is an infinity, raise the corresponding saturation flag.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp2,0xFF000000
	feq.vs vAbsA,temp2
	cfvu temp0,vcond
	or saturationFlags,temp0

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vA is a NaN, raise the corresponding flag in fltflags_invalid32 and clear
the corresponding saturation flag.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	fltu.sv temp2,vAbsA
	cfvu temp1,vcond
	lw temp0,fltflags_invalid32
	or temp0,temp1
	sw temp0,fltflags_invalid32
	not temp1
	and saturationFlags,temp1

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Return.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	j ra

	.end tor_conv_gf_gx32s

