/*
** $Header: /u/spert/src/fltvec/tor/RCS/tor_conv_gx32_gf.S,v 1.4 1995/07/05 22:36:36 davidj Exp $
*/

#include "tor/tordef.h"
#include "m_regdef.h"

	.globl tor_conv_gx32_gf
#ifdef DECL_tor_conv_gx32_gf 
void tor_conv_gx32_gf(void); 
#endif
	.globl conv_gx16_gf

#define vA va0			/* Input A				*/
#define aExp t2			/* Exponent of A			*/
#define vZ vv2			/* Output Z				*/

#define vZSign vt0		/* Sign of result			*/
#define vRawZSignif vt2		/* Raw significand of result		*/
#define vCount16 vt4		/* vCountLeadingZerosA & 0x10		*/
#define vCount8 vt5		/* vCountLeadingZerosA & 0x08		*/
#define vCountLeadingZerosA vt1	/* Number of leading zeros of vA	*/
#define vHighIndex vt3		/* Index into low zeros-count table	*/
#define vLowIndex vt4		/* Index into high zeros-count table	*/
#define vZExp vt1		/* Under-biased exponent of result	*/
#define vZSignif vv2		/* Significand of result		*/
#define vSignifScaledBack vt3	/* vSubnormalSignif scaled back to	*/
				/*     correspond with vRawZSignif	*/
#define vSubnormalSignif vt4	/* Significand of result if subnormal	*/
#define vCountB vt5		/* Amount to shift down vRawZSignif for	*/
				/*     subnormal result			*/
#define vNotUnderflowB vtt	/* Flag for result does not underflow	*/
#define vNotUnderflow vt5	/* Flag for result does not underflow	*/
#define vOverflow1 vt4		/* First flag for result overflows	*/
#define vOverflow2Not vt5	/* Second flag for result overflows	*/

#define temp0 v0
#define temp1 v1
#define temp2 t4
#define temp3 t5
#define config0 t6
#define config1 t7
#define config2 t8
#define config3 t9

	.ent tor_conv_gx32_gf

/*
-------------------------------------------------------------------------------
Routine:  tor_conv_gx32_gf

Input:
	vlr
	va0	vA
	t2	aExp

Output:
	vv2	vZ

Not preserved:
	vtt
	vv3
	va0--va3
	vt0--vt5
	v0--v1
	t4--t9

Function:

The 32-bit fixed-point vector vA is element-wise converted to a single-
precision floating-point vector and the result returned in vZ.  The exponent of
the vA vector is given by aExp.  The result is rounded to nearest/even.  The
vlr register controls the elements of vZ actually updated.  The floating-point
format conforms to the IEEE Standard for Binary Floating-point Arithmetic, and
the conversion is performed in a manner consistent with that standard.

Exceptions for any of the element-wise conversions are signalled by raising
element-wise flags in the global variables fltflags_overflow32 and
fltflags_underflow32.  No traps are taken for any exceptions.  Underflow is
never signalled for results that are not subnormal.

-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
Routine:  conv_gx16_gf

Input:
	vlr
	va0	vA
	t2	aExp

Output:
	vv2	vZ

Not preserved:
	vtt
	vv3
	va0--va3
	vt0--vt5
	v0--v1
	t4--t9

Function:

The 16-bit fixed-point vector vA is element-wise converted to a single-
precision floating-point vector and the result returned in vZ.  The exponent of
the vA vector is given by aExp.  If the result is a subnormal, it is rounded to
nearest/even.  The vlr register controls the elements of vZ actually updated.
The floating-point format conforms to the IEEE Standard for Binary Floating-
point Arithmetic, and the conversion is performed in a manner consistent with
that standard.

Exceptions for any of the element-wise conversions are signalled by raising
element-wise flags in the global variables fltflags_overflow32 and
fltflags_underflow32.  No traps are taken for any exceptions.  Underflow is
never signalled for results that are not subnormal.

-------------------------------------------------------------------------------
*/

conv_gx16_gf:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	va0	vA
|	t2	aExp
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Convert 16-bit fixed-point to 32-bit fixed-point and then do 32-bit-to-
single-precision conversion.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,16
	sllv.vs vA,vA,temp0

tor_conv_gx32_gf:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	va0	vA
|	vt0	vZSign
|	t2	aExp
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set vZSign according to the sign of vA, and then take the absolute value of vA.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	slt.vv vZSign,vA,vzero
	li config0,(FX_CLIP_PASS|FX_GEZ)
	fxsub.vv vA,vzero,vA,config0

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	va0	vA
|	vt0	vZSign
|	vt1	vCountLeadingZerosA
|	vt2	vRawZSignif
|	vt3	vHighIndex
|	vt4	vCount16,vLowIndex
|	vt5	vCount8
|	t2	aExp
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Set vCountLeadingZerosA to the number of leading zeros of vA.  Shift up to
normalize in vRawZSignif.
    (NOTE:  This block of code could all be replaced by a count-leading-zeros
instruction followed by a shift.  See countlz_gi32_gi32.S for more refined
commentary.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,(1<<16)
	sltu.vs vCount16,vA,temp0
	li temp1,4
	sllv.vs vCount16,vCount16,temp1
	sllv.vv vRawZSignif,vA,vCount16
	li temp2,(1<<24)
	sltu.vs vCount8,vRawZSignif,temp2
	li temp3,3
	sllv.vs vCount8,vCount8,temp3
	sllv.vv vRawZSignif,vRawZSignif,vCount8
	addu.vv vCount8,vCount16,vCount8
	li config1,(FX_SHLZA|FX_A_AND_B|(23*FX_SHAM)|FX_NOROUND|FX_CLIP_PASS)
	li config2,(FX_SHLV|(28*FX_SHAM)|FX_NOROUND|FX_CLIP_PASS|FX_GEZ)
	li temp0,(7<<25)
	fxadd.vs vLowIndex,vRawZSignif,temp0,config1
	li temp1,0x76554444
	fxadd.sv vCountLeadingZerosA,temp1,vLowIndex,config2
	li temp2,(7<<28)
	addu config3,config1,(3*FX_SHAM)
	fxadd.vs vHighIndex,vRawZSignif,temp2,config3
	li temp3,0xF3221111
	fxadd.sv vCountLeadingZerosA,temp3,vHighIndex,config2
	cmvltz.vv vCountLeadingZerosA,vRawZSignif,vzero
	sllv.vv vRawZSignif,vRawZSignif,vCountLeadingZerosA
	addu.vv vCountLeadingZerosA,vCount8,vCountLeadingZerosA

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vt0	vZSign
|	vt1	vCountLeadingZerosA,vZExp
|	vt2	vRawZSignif
|	t2	aExp
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Calculate the "under-biased" result exponent vZExp from vCountLeadingZerosA
and aExp.  (The exponent is "under-biased" because the integral portion of the
significand will be added to it when the result is packed.  In most cases, a 1
will be added to the exponent during packing.  However, if a carry occurred
during rounding of the significand, a value of 2 will be added to the exponent
during packing.)  If the result significand is exactly zero, replace the result
exponent by zero.  (If the significand is zero, the value added to the exponent
during packing will be zero.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,0x7E
	subu.sv vZExp,temp0,vCountLeadingZerosA
	sadd.vs vZExp,vZExp,aExp
	cmveqz.vv vZExp,vRawZSignif,vzero

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
|	vv2	vZSignif,vZ
|	t2	aExp
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Form the result significand by shifting down the raw result significand,
rounding to nearest/even.
    Subnormal results are (mis)treated as though they were normal.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config0,(FX_BU|(8*FX_SHAM)|FX_SHRL|FX_RNEARESTEVEN|FX_CLIP_PASS)
	fxadd.vv vZSignif,vzero,vRawZSignif,config0

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vZExp is negative, the result is a subnormal; otherwise, if vZExp is 0xFD or
more, the result might have overflowed.  If either case occurs for any element
of the result, jump to a special case handler.
    (Although the result has not actually overflowed in most cases in which
vZExp equals 0xFD, overflow does occur when vZExp is 0xFD and vRawZSignif >=
0xFFFFFF80.  Since a zero significand forced a zero exponent earlier, an exact
zero result will not cause a jump to the special case handler.  In all other
cases, if vZExp is zero, the biased exponent of the result will be at least 1.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	ctvu zero,vcond
	li temp2,0xFC
	fltu.sv temp2,vZExp
	cfvu temp3,vcond
	bnez temp3,specialCases

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Add the integral portion of the significand to the under-biased exponent, and
pack the sign, biased exponent, and significand of the result.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config1,((31*FX_SHLAM)|FX_SEPSHAM|FX_NOROUND)
	fxadd.vv vZ,vZSign,vZSignif,config1
	li config2,((23*FX_SHLAM)|FX_SEPSHAM|FX_NOROUND)
	fxadd.vv vZ,vZExp,vZ,config2

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Return.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	j ra

specialCases:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
|	vt3	vSignifScaledBack
|	vt4	vSubnormalSignif
|	vt5	vCountB
|	vv2	vZSignif
|	t2	aExp
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vZExp is negative, set vSubnormalSignif to the result subnormal significand
by shifting down vRawZSignif, rounding to nearest/even.  Set vSignifScaledBack
to vSubnormalSignif scaled back to correspond with vRawZSignif.  If vZExp is
nonnegative, vSubnormalSignif and vSignifScaledBack are trash.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	negu.vv vCountB,vZExp
	li temp0,31
	li config0,(FX_AU|FX_LURES|FX_CLIP_PASS|FX_GTZ)
	fxsub.vs vCountB,vCountB,temp0,config0
	li config1,(FX_AU|FX_SHRL|FX_SHRV|FX_RODD|FX_CLIP_PASS)
	fxadd.vv vSubnormalSignif,vRawZSignif,vCountB,config1
	li config2,(FX_BU|(8*FX_SHAM)|FX_RNEARESTEVEN|FX_CLIP_PASS)
	fxadd.vv vSubnormalSignif,vzero,vSubnormalSignif,config2
	sllv.vv vSignifScaledBack,vSubnormalSignif,vCountB
	li temp1,8
	sllv.vs vSignifScaledBack,vSignifScaledBack,temp1

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vNotUnderflowB
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
|	vt3	vSignifScaledBack
|	vt4	vSubnormalSignif
|	vt5	vNotUnderflow
|	vv2	vZSignif
|	t2	aExp
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If the result underflows, raise the corresponding flag in fltflags_underflow32.
Underflow occurs if:  (1) vZExp is negative, and (2) vSubnormalSignif remained
a subnormal even after rounding, and (3) some nonzero bits of vRawZSignif were
rounded off.  (vSubnormalSignif might round up to the smallest non-subnormal
number, in which case underflow is not signalled.  Nonzero bits were rounded
off exactly if vSignifScaledBack is not equal to vRawZSignif.  temp1 is just a
convenient nonzero value.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	seq.vv vNotUnderflow,vSignifScaledBack,vRawZSignif
	li temp0,23
	srlv.vs vNotUnderflowB,vSubnormalSignif,temp0
	cmvnez.vv vNotUnderflow,vNotUnderflowB,vNotUnderflowB
	cmvgez.vs vNotUnderflow,vZExp,temp1
	ctvu zero,vcond
	feqz.v vNotUnderflow
	cfvu temp2,vcond
	lw temp3,fltflags_underflow32
	or temp3,temp2
	sw temp3,fltflags_underflow32

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vZExp is negative, set vZSignif to the (probably) subnormal result
significand and set vZExp to zero.  (For subnormal results, the exponent and
significand are given entirely by vZSignif.  The exponent is normally zero, but
will be 1 if vRawZSignif >= 0xFFFFFF80.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	cmvltz.vv vZSignif,vZExp,vSubnormalSignif
	cmvltz.vv vZExp,vZExp,vzero

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vt0	vZSign
|	vt1	vZExp
|	vt4	vOverflow1
|	vt5	vOverflow2Not
|	vv2	vZSignif, vZ
|	t2	aExp
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Add the integral portion of the significand to the under-biased exponent, and
pack the exponent and significand of the result.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config0,((23*FX_SHLAM)|FX_SEPSHAM|FX_NOROUND)
	fxadd.vv vZ,vZExp,vZSignif,config0

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If the result overflows, raise the corresponding flag in fltflags_overflow32.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp1,0xFF
	sltu.sv vOverflow1,temp1,vZExp
	li temp2,(0xFF<<23)
	sltu.vs vOverflow2Not,vZ,temp2
	flt.vv vOverflow1,vOverflow2Not
	cfvu temp3,vcond
	lw temp0,fltflags_overflow32
	not temp3
	or temp0,temp3
	sw temp0,fltflags_overflow32

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If the result overflows, replace it by an infinity.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	cmvnez.vs vZ,vOverflow1,temp2
	cmveqz.vs vZ,vOverflow2Not,temp2

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Pack the sign, biased exponent, and significand of the result.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config1,((31*FX_SHLAM)|FX_SEPSHAM|FX_NOROUND)
	fxadd.vv vZ,vZSign,vZ,config1

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Return.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	j ra

	.end tor_conv_gx32_gf

