/*
** $Header: /u/spert/src/fltvec/tor/RCS/tor_fltvec_special.S,v 1.5 1995/07/08 00:54:01 davidj Exp $
*/

#include "tor/tordef.h"
#include "m_regdef.h"

	.globl tor_fltv_singlespec
	.globl tor_fltv_singlerange

	.extern tor_fltv_flags_underflow32,4
	.extern tor_fltv_flags_overflow32,4
	.extern tor_fltv_flags_divbyzero32,4
	.extern tor_fltv_flags_invalied32,4

#define vInputsCode vtt
#define vZSign vt0
#define vZExp vt1
#define vRawZSignif vt2
#define vZSignif vv2
#define vZ vv2

#define vInvalidOperation vt3
#define vSignifScaledBack vt3
#define vSubnormalSignif vt4
#define vCount vt5
#define vNotUnderflow2 vtt
#define vNotUnderflow vt5
#define vCarry vt5
#define vOverflow vt5

#define temp0 v0
#define temp1 v1
#define temp2 t4
#define temp3 t5
#define config0 t6
#define config1 t7
#define config2 t8
#define config3 t9

	.ent tor_fltv_singlespec

/*
-------------------------------------------------------------------------------

Routine:  tor_fltv_singlespec

Input:
	vtt	vInputsCode
	vt0	vZSign
	vt1	vZExp
	vt2	vRawZSignif
	vv2	vZSignif

Output:
	vv2	vZ

Not preserved:
	vtt
	vv3
	vt0--vt5
	v0--v1
	t4--t9

Function:

The vInputsCode

The raw floating-point values given by the combination of the vectors vZSign,
vZExp, vRawZSignif, and vZSignif are element-wise checked for single-precision
range violations, and substitution of infinities and correctly rounded
subnormal numbers are made as appropriate.  The result values are returned as
ordinary single-precision floating-point numbers in vector vZ.  The vlr
register controls the elements of vZ actually updated.  The exact operation
performed by this function can differ on an element-wise basis between two
options:

Distill the types of the original inputs into a single value in vInputsCode.
Set vInputsCode to between 3 and 6 inclusive if either input is a NaN; else set
vInputsCode to 0 or 1 if either input is an infinity; else set vInputsCode to
exactly 2.  (This distillation is done by first setting vACode to either 1, 0,
or 3, according to whether the original A input is normal, an infinity, or a
NaN, respectively.  The same is done with vBCode, and then vACode and vBCode
are added to form vInputsCode.  This value is used by the special case handler
called below.)


NaNs and infinities must have been added or multiplied as though they were normal

-------------------------------------------------------------------------------

-------------------------------------------------------------------------------

Routine:  tor_fltv_singlerange

Input:
	vt0	vZSign
	vt1	vZExp
	vt2	vRawZSignif
	vv2	vZSignif

Output:
	vv2	vZ

Not preserved:
	vtt
	vv3
	vt0--vt5
	v0--v1
	t4--t9

Function:

The raw floating-point values given by the combination of the vectors vZSign,
vZExp, vRawZSignif, and vZSignif are element-wise checked for single-precision
range violations, and substitution of infinities and correctly rounded
subnormal numbers are made as appropriate.  The result values are returned as
ordinary single-precision floating-point numbers in vector vZ.  The vlr
register controls the elements of vZ actually updated.  The exact operation
performed by this function can differ on an element-wise basis between two
options:

In the first option, vRawZSignif holds the raw, rounded-odd significand; vZExp
holds the exponent, and vZSign holds the sign.  vRawZSignif must have a leading
1 bit in the most-significant bit position.  vZSignif must contain the same
significand as vRawZSignif, rounded to 24 places.  vZExp must be one less than
the raw biased exponent (called under-biased).

For this option, the raw result is converted to a final single-precision
floating-point result according to the requirements of the IEEE Standard for
Binary Floating-point Arithmetic.  If the result overflows an infinity is
substituted and the corresponding flag in the global variable
tor_fltv_flags_overflow32 is raised.  If the result is too small to be represented as
a normal floating-point number, the correctly rounded subnormal result is
substituted.  If the result underflows, the corresponding flag in the global
variable tor_fltv_flags_underflow32 is raised.  Underflow is never signalled for
results that are not subnormal or are exact.

In the second option, vZSignif holds an entire floating-point value.  vZExp
must be zero.  In this option, the result returned is vZSignif with its sign
XOR-ed with vZSign (0 or 1).  No exceptions are signalled.  vRawZSignif is
ignored.  This option should be used to pass infinity and NaN results through
to vZ.

-------------------------------------------------------------------------------
*/

tor_fltv_singlespec:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vInputsCode
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
|	vt3	vInvalidOperation
|	vv2	vZSignif
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Adjust vInputsCode so that a positive (nonzero) value means that one or both of
the original inputs was a NaN, or else a negative value means that one or both
of the original inputs was an infinity, while a zero value means that both
inputs were non-NaN and finite.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,2
	subu.vs vInputsCode,vInputsCode,temp0

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If neither original input was a NaN, at least one input was an infinity, and
the raw result significand is zero, set vInvalidOperation nonzero to indicate
that an invalid operation has occurred.  For an addition, if one of the inputs
was an infinity, the raw result significand will be zero only if an addition of
infinities of opposite sign occurred.  For a multiplication, if one of the
inputs was an infinity, the raw result significand will be zero only if the
other input was a zero.
    If an invalid operation occurred, set vInputsCode as though one of the
original inputs was a NaN.  (temp0 is just a convenient positive value.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	sltz.vv vInvalidOperation,vInputsCode
	cmvnez.vv vInvalidOperation,vRawZSignif,vzero
	cmvnez.vs vInputsCode,vInvalidOperation,temp0

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If one of the original inputs was a NaN or if an invalid operation occurred,
set vZSignif to a default MIPS single-precision NaN (0x7FBFFFFF) and set vZExp
and vZSign to zero.  Else if one of the inputs was an infinity, set vZSignif to
an infinity and set vZExp to zero.  Setting vZExp to zero indicates to
tor_fltv_singlerange below that vZSignif and vZSign together contain the
entire floating-point result, and that range violations should not be
signalled.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp1,0x7F800000
	cmvltz.vs vZSignif,vInputsCode,temp1
	li temp2,0x7FBFFFFF
	cmvgtz.vs vZSignif,vInputsCode,temp2
	cmvnez.vv vZExp,vInputsCode,vzero
	cmvgtz.vv vZSign,vInputsCode,vzero

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If an invalid operation occurred, raise the corresponding flag in
tor_fltv_flags_invalid32.
    Fall into the tor_fltv_singlerange function below.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	ctvu zero,vcond
	fnez.v vInvalidOperation
	cfvu temp0,vcond
	lw temp3,tor_fltv_flags_invalid32
	or temp3,temp0
	sw temp3,tor_fltv_flags_invalid32

tor_fltv_singlerange:

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
|	vt3	vSignifScaledBack
|	vt4	vSubnormalSignif
|	vt5	vCount
|	vv2	vZSignif
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vZExp is negative, set vSubnormalSignif to the result subnormal significand
by shifting down vRawZSignif, rounding to nearest/even.  Set vSignifScaledBack
to vSubnormalSignif scaled back to correspond with vRawZSignif.  If vZExp is
nonnegative, vSubnormalSignif and vSignifScaledBack are trash.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	negu.vv vCount,vZExp
	li temp0,31
	li config0,(FX_AU|FX_LURES|FX_CLIP_PASS|FX_GTZ)
	fxsub.vs vCount,vCount,temp0,config0
	li config1,(FX_AU|FX_SHRL|FX_SHRV|FX_RODD|FX_CLIP_PASS)
	fxadd.vv vSubnormalSignif,vRawZSignif,vCount,config1
	li config2,(FX_BU|(8*FX_SHLAM)|FX_RNEARESTEVEN|FX_CLIP_PASS)
	fxadd.vv vSubnormalSignif,vzero,vSubnormalSignif,config2
	sllv.vv vSignifScaledBack,vSubnormalSignif,vCount
	li temp1,8
	sllv.vs vSignifScaledBack,vSignifScaledBack,temp1

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vtt	vNotUnderflow2
|	vt0	vZSign
|	vt1	vZExp
|	vt2	vRawZSignif
|	vt3	vSignifScaledBack
|	vt4	vSubnormalSignif
|	vt5	vNotUnderflow
|	vv2	vZSignif
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If the result underflows, raise the corresponding flag in tor_fltv_flags_underflow32.
Underflow occurs if:  (1) vZExp is negative, (2) vSubnormalSignif remained a
subnormal even after rounding, and (3) some nonzero bits of vRawZSignif were
rounded off.  vSubnormalSignif might round up to the smallest non-subnormal
number, in which case underflow is not signalled.  (Nonzero bits were rounded
off exactly if vSignifScaledBack is not equal to vRawZSignif.  temp1 is just a
convenient nonzero value.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	seq.vv vNotUnderflow,vSignifScaledBack,vRawZSignif
	li temp0,23
	srlv.vs vNotUnderflow2,vSubnormalSignif,temp0
	cmvnez.vv vNotUnderflow,vNotUnderflow2,vNotUnderflow2
	cmvgez.vs vNotUnderflow,vZExp,temp1
	ctvu zero,vcond
	feqz.v vNotUnderflow
	cfvu temp2,vcond
	lw temp3,tor_fltv_flags_underflow32
	or temp3,temp2
	sw temp3,tor_fltv_flags_underflow32

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vZExp is negative, set vZSignif to the (probably) subnormal result
significand and set vZExp to zero.  (For subnormal results, the exponent and
significand are given entirely by vZSignif.  The exponent is normally zero, but
will be 1 if vRawZSignif >= 0xFFFFFF80.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	cmvltz.vv vZSignif,vZExp,vSubnormalSignif
	cmvltz.vv vZExp,vZExp,vzero

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vt0	vZSign
|	vt1	vZExp
|	vt5	vCarry
|	vv2	vZSignif, vZ
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vZSignif holds the entire final result, set vZ to vZSignif.  (If vZSignif
holds the entire result, vZExp is zero.)  Otherwise, add the integral portion
of the significand to the under-biased exponent to form the final exponent, and
pack the exponent and significand of the result.  If vZExp is nonzero, update
it to reflect the final exponent value; otherwise, if vZExp is zero, leave it
zero.  vZExp is used below to determine if overflow occurred.
    (vZExp is the only indication of whether vZSignif holds the entire result.
If vZExp is zero, vZSignif might hold the entire result; overflow should not be
signalled if it does.  If vZExp is zero but vZSignif does not hold the entire
result, the result could not overflow anyway.)
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,23
	srlv.vs vCarry,vZSignif,temp0
	cmveqz.vv vCarry,vZExp,vzero
	li config3,((23*FX_SHLAM)|FX_SEPSHAM|FX_NOROUND)
	fxadd.vv vZ,vZExp,vZSignif,config3
	sadd.vv vZExp,vZExp,vCarry

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
|	vt0	vZSign
|	vt1	vZExp
|	vt5	vOverflow
|	vv2	vZ
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If the result overflows, raise the corresponding flag in tor_fltv_flags_overflow32.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp1,0xFE
	slt.sv vOverflow,temp1,vZExp
	fnez.v vOverflow
	cfvu temp2,vcond
	lw temp3,tor_fltv_flags_overflow32
	or temp3,temp2
	sw temp3,tor_fltv_flags_overflow32

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If the result overflows, replace it by an infinity.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li temp0,(0xFF<<23)
	cmvnez.vs vZ,vOverflow,temp0

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
If vZSignif (and thus vZ) holds the entire result, invert the sign of vZ
according to vZSign.  Otherwise, simply pack the sign and significand of the
result.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	li config0,((31*FX_SHLAM)|FX_SEPSHAM|FX_NOROUND)
	fxadd.vv vZ,vZSign,vZ,config0

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
Return.
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
	j ra

	.end tor_fltv_singlespec

